import matplotlib.pyplot as plt
import numpy as np
import powerlaw


def get_dist(data):
    data = sorted(data)
    xdata = []
    ydata = []
    x = data[0]
    y = 1
    i = 1
    l = len(data)
    while(i<l):
        xnew = data[i]
        if xnew == x:
            y += 1
        else:
            xdata.append(x)
            ydata.append(y/l)
            x = xnew
            y = 1
        i += 1  
    xdata.append(x)
    ydata.append(y/l)
    return xdata, ydata


def fit_data(data, xmin=None, xmax=None, compare=False, plot=False, discrete=True, output_file='', verbose=False):
    ''' fit data to a power law

        Args:
            data            is the distribution to fit, can be either a list or a dictionary
                            if a list, is interpreted as the list of all outcomes (e.g., [0,1,2,1,1] means pr(0)=pr(2)=0.2, pr(1)=0.6)
                            if a dictionary, items are interpreted as outcome:number_of_occurrences (e.g., {0:1, 1:3, 2:1})
            xmin
            xmax
            compare         another distribution to which the power law fit can be compared; if True or invalid, exponential is used 
            plot            specifies whether the fit must be plotted (both pdf fit and ccdf fit are plotted)
            output_file     the output file for the plot; if empty a default name is used
    '''
    if isinstance(data,dict):
        d = []
        for k,v in data.items():
            d += [k]*v
        data = np.array(d)
    else:
        data = np.array(data)

    fit = powerlaw.Fit(data=data, discrete=discrete, xmin=xmin, xmax=xmax, linear_bins=True)
    if verbose:
        print('Power-law best fit to data:')
        print('\t alpha = {}'.format(fit.power_law.alpha))
        print('\t sigma = {}'.format(fit.power_law.sigma))
        print('\t xmin = {}'.format(fit.power_law.xmin))
        print('\t xmax = {}'.format(fit.power_law.xmax))
    if compare:
        if compare == True:
            R, p = fit.distribution_compare('power_law', 'exponential', normalized_ratio = True)
        else:
            try:
                R, p = fit.distribution_compare('power_law', compare, normalized_ratio=True)
            except:
                print('the argument to option "compare" is not a valid distribution, will use exponential')
                R, p = fit.distribution_compare('power_law', 'exponential', normalized_ratio = True)
        if verbose:
            print('Normalized loglikelihood ratio: {}'.format(R))
            print('Significance value: {}'.format(p))
    if plot:
        fig, ax = plt.subplots(nrows=1, ncols=1,figsize=(10,10))
        fit.plot_pdf(color= 'b', ax=ax, label='data pdf')
        fit.power_law.plot_pdf(color= 'b',linestyle='--',label='fit pdf', ax=ax)
        fit.plot_ccdf(color='r', ax=ax, label='data ccdf')
        fit.power_law.plot_ccdf(color='r', linestyle='--', ax=ax, label='fit ccdf')
        plt.legend()
        if not output_file:
            output_file = 'power_law_fit.png'
        fig.savefig(output_file)
    return fit


def get_fit_dist(alpha, xmin, xmax, scale):
    xfit = np.linspace(xmin,xmax,num=100)
    beta = (alpha-1)/(xmin**(1-alpha)-xmax**(1-alpha)) * scale
    yfit = beta*xfit**(-alpha)
    return xfit, yfit


def fibonacci_bins(x, s=None):
    t = False
    if s == None:
        s = 1
    else:
        t = True
    edges = [1,1]
    while edges[-1]+s-1 < x:
        edges.append(edges[-2]+edges[-1])
    if t:
        edges = [e+s-1 for e in edges]
    return [(edges[i],edges[i+1]) for i in range(1,len(edges)-1)]


def fibonacci_hist_cont(data, num):

    l = len(data)
    data = sorted(data)
    xmin = data[0]
    i = 0
    if xmin == 0:
        i = 1
        while(data[i]==0):
            i += 1
        xmin = data[i]
        data = data[i:]
        i /= l
    xmax = data[-1]
    
    binmin = 0
    binmax = num-1    
    bins = fibonacci_bins(binmax,binmin)
    f = (xmax-xmin)/num
    new_bins = [(xmin+b[0]*f,xmin+b[1]*f) for b in bins]
    
    hist = [0]*len(new_bins)
    j = 0
    for d in data:
        if new_bins[j][0] <= d < new_bins[j][1]:
            hist[j] += 1
        else:
            j += 1
    if i>0:
        xdata = [0.0]
        ydata = [i]
    else:
        xdata = []
        ydata = []
        
    xdata += [(b[0]+b[1])/2 for b in new_bins] 
    ydata += [h/(l*(bins[j][1]-bins[j][0])) for j,h in enumerate(hist)]

    return xdata, ydata