import sys
import time
import random
import pandas as pd
import pickle
import matplotlib.pyplot as plt
import numpy as np
from itertools import combinations, chain
import powerlaw
#################################
import igraph
from igraph import *
#################################
from anodelib import aNodes
from datamanager import fibonacci_bins, fibonacci_hist_cont, fit_data, get_dist, get_fit_dist


class graphmanager():
    
    def __init__(self, gdict={}, vkey='', vattr=[], eattr=[]):
        self.gdict = gdict
        for name in self.gdict:
            self.gdict[name]['igraph'] = None
        self.vkey = vkey
        self.vattr = vattr
        self.eattr = eattr
        self.filename = ''
        
        
    def load(self, filename):
        with open(filename, 'rb') as fin:
            tmp_dict = pickle.load(fin)        
        self.__dict__.update(tmp_dict)
        self.filename = filename


    def save(self, filename=''):
        if filename:
            self.filename = filename
        with open(self.filename, 'wb') as fout:
            pickle.dump(self.__dict__, fout)
        

    def load_all_igraph_objects(self):
        for name in self.gdict:
            self.load_igraph_object(name)


    def dump_all_igraph_objects(self):
        for name in self.gdict:
            self.dump_igraph_object(name)
            
            
    def read_all_graphs(self):
        for name in self.gdict:
            self.read_graph(name)
    
        
    def load_igraph_object(self, graph_name):
        
        try:
            print('Loading igraph object for graph {}...'.format(graph_name))
            stime = time.time()
            self.gdict[graph_name]['igraph'] = Graph.Read_Pickle(self.gdict[graph_name]['pickle_file'])
            print('Time to load graph: {}'.format(time.strftime("%H:%M:%S", time.gmtime(time.time() - stime))))            
        except (IOError,ValueError):
            print('Loading failed!')
            sys.exit()
        
        
    def dump_igraph_object(self, graph_name):    

        print('Dumping igraph object for graph {}...'.format(graph_name))
        stime = time.time()
        self.gdict[graph_name]['igraph'].write_pickle(self.gdict[graph_name]['pickle_file'])
        print('Time to dump: {}'.format(time.strftime("%H:%M:%S", time.gmtime(time.time() - stime))))
        
    
    def read_graph(self, graph_name):
        
        edges_file = self.gdict[graph_name]['edges_file']
        nodes_file = self.gdict[graph_name]['nodes_file']
        
        print('Reading edges from {}...'.format(edges_file))
        stime = time.time()
        try:
            graph = aNodes.getEdges(edges_file, self.eattr, self.vkey)
            print('Time to read edges: {}'.format(time.strftime("%H:%M:%S", time.gmtime(time.time() - stime))))
        except (IOError,ValueError):
            print('Edges reading failed!')
            sys.exit()
                
        print('Reading nodes from {}...'.format(nodes_file))
        stime = time.time()
        try:
            graph.addAttributesFromFile(nodes_file, self.vattr)
            print('Time to read nodes: {}'.format(time.strftime("%H:%M:%S", time.gmtime(time.time() - stime))))
        except (IOError,ValueError):
            print('Nodes reading failed!')
            sys.exit()

        print('Building igraph object...')
        stime = time.time()
        g = graph.iGraphObject() # load_vectors=bool(vectors_file))
        print('Time to build: {}'.format(time.strftime("%H:%M:%S", time.gmtime(time.time() - stime))))

        if not g.is_simple():
            print('Flattening graph...')
            stime = time.time()
            g.simplify(combine_edges='ignore')
            print('Time to flatten: {}'.format(time.strftime("%H:%M:%S", time.gmtime(time.time() - stime))))
        
        if not g.is_connected(mode='WEAK'):
            print('Finding giant component')
            stime = time.time()
            g = g.components(mode='WEAK').giant()
            print('Time to find giant component: {}'.format(time.strftime("%H:%M:%S", time.gmtime(time.time() - stime))))
                
        self.gdict[graph_name]['igraph'] = g
        
        
    def get_intersection(self, graph_names=None, select_by='url'):

        if not graph_names:
            graph_names = list(self.gdict.keys())

        graphs = [self.gdict[name]['igraph'] for name in graph_names]

        values = []
        for graph in graphs:
            values.append(set(graph.vs[select_by]))

        intersection = set.intersection(*values)
        ginter = []
        for graph in graphs:
            to_keep = [v.index for v in graph.vs if v.attributes().get(select_by) in intersection]
            ginter.append(graph.subgraph(to_keep))

        edge_dicts = [{(graph.vs[e.source][select_by],graph.vs[e.target][select_by]):e.index for e in graph.es} for graph in graphs] 
        edge_sets = [set(ed.keys()) for ed in edge_dicts] 
        common_edges = set.intersection(*edge_sets)
        common_edges_indices = [edge_dicts[0][e] for e in common_edges]

        g = graphs[0].subgraph_edges(common_edges_indices)   
        g = g.components(mode='WEAK').giant()

        if set(graph_names) != set(self.gdict.keys()):
            key = 'intersection_{}'.format('_'.join(graph_names))
        else:
            key = 'intersection'

        self.gdict[key] = {}
        self.gdict[key]['igraph'] = g
        name_start = self.gdict[graph_names[0]]['basename'].find(graph_names[0])
        basename = self.gdict[graph_names[0]]['basename'][:name_start] + key + '.services'
        self.gdict[key]['basename'] = basename
        self.gdict[key]['pickle_file'] = basename + '.igraph.pickle'

        
####################### METRICS ########################
        
    @classmethod
    def global_efficiency(cls, graph, mode='BOTH'):
        '''compute the in- and/or out- global efficiency of the input graph
        '''
        if mode == 'IN':
            mode = -1
        elif mode == 'OUT':
            mode = +1
        else:
            mode = 0
        effin  = 0.0
        effout = 0.0
        vcount = graph.vcount()
        # c = 0
        for v in range(vcount): 
            pathlengths = graph.shortest_paths_dijkstra(source=v)
            if mode >= 0:
                pathlengthsout = pathlengths[0][v+1:] # remove the zero: pl[0][:v] + pl[0][v+1:]  
                for l in pathlengthsout:
                    effout += 1/l
            if mode <= 0:
                pathlengthsin  = pathlengths[0][:v]
                for l in pathlengthsin:
                    effin += 1/l
        if mode > 0:
            return effout/(vcount*(vcount-1))
        elif mode < 0:
            return effin/(vcount*(vcount-1))
        else:
            return effin/(vcount*(vcount-1)), effout/(vcount*(vcount-1))     

        
    @classmethod        
    def transitivity(cls, graph):
        '''Compute two variants of the global transitivity for directed graphs
        variant "cycles": probability that a cycle is closed, that is, given C->A and A->B, check if B->C
        variant "any"   : probability that a fork is closed, that is, given A->B and A->C, we check if B->C and/or C->B

        vertices with zero denominator (no triplets) are assigned 0 transitivity
        '''

        vcount = graph.vcount()
        triangles = {'cycles':[0.0 for _ in range(vcount)], 'any':[0.0 for _ in range(vcount)]}
        triplets  = {'cycles':[0.0 for _ in range(vcount)], 'any':[0.0 for _ in range(vcount)]}
        for v in graph.vs:
            vi = v.index
            for wout in v.neighbors('out'):
                for win in v.neighbors('in'):
                    triplets['cycles'][vi] += 1
                    if graph.are_connected(wout,win):
                        triangles['cycles'][vi] += 1
                for uout in v.neighbors('out'):
                    if uout != wout:
                        triplets['any'][vi] += 1
                        e1 = e2 = 0
                        if graph.are_connected(wout,uout):
                            e1 = 1
                        if graph.are_connected(uout,wout):
                            e2 = 1
                        s = e1+e2
                        m = e1*e2
                        triangles['any'][vi] += s-m
        tglob = {'cycles':0.0, 'any':0.0}
        for key in ['cycles', 'any']:
            denom = sum(t for t in triplets[key])
            if denom != 0:
                tglob[key] = sum(t for t in triangles[key])/denom

        return tglob['cycles'], tglob['any']
    
    
    @classmethod    
    def compute_metrics(cls, graph):

        metrics = {}

        vcount          = graph.vcount()
        ecount          = graph.ecount()
        indegs          = graph.indegree()
        outdegs         = graph.outdegree()
        pagerank        = graph.pagerank()
        betweenness     = graph.betweenness()
        maxindeg        = max(indegs)
        maxoutdeg       = max(outdegs)
        trans1, trans2  = cls.transitivity(graph)
        effin, effout   = cls.global_efficiency(graph)

        metrics['vcount']          = vcount
        metrics['ecount']          = ecount
        metrics['density']         = ecount/(vcount*(vcount-1))
        metrics['minindeg']        = min(indegs)
        metrics['maxindeg']        = maxindeg
        metrics['minoutdeg']       = min(outdegs)
        metrics['maxoutdeg']       = maxoutdeg
        metrics['avgdeg']          = sum(indegs)/vcount
        metrics['assortativity']   = graph.assortativity_degree()
        metrics['indegcen']        = (vcount*maxindeg-sum(indegs))/(vcount-1)**2
        metrics['outdegcen']       = (vcount*maxoutdeg-sum(outdegs))/(vcount-1)**2
        metrics['clustering']      = graph.transitivity_undirected()
        metrics['trans1']          = trans1
        metrics['trans2']          = trans2
        metrics['diameter']        = graph.diameter()
        metrics['radiusin']        = graph.radius('IN')
        metrics['radiusout']       = graph.radius('OUT')
        metrics['avgsplen']        = graph.average_path_length()
        metrics['efficiencyin']    = effin
        metrics['efficiencyout']   = effout
        metrics['centralities']    = {'indegree':indegs, 'outdegree':outdegs, 'pagerank':pagerank, 'betweenness':betweenness}

        return metrics
    
    
    def get_metrics(self, graph_names=None):
        '''get metrics for all input graphs'''

        if not graph_names:
            graph_names = list(self.gdict.keys())
    
        for name in graph_names:
            metrics = self.compute_metrics(self.gdict[name]['igraph'])
            self.gdict[name]['centralities'] = metrics['centralities']
            del(metrics['centralities'])
            self.gdict[name]['global_metrics'] = metrics
            
            
    def global_metrics_table(self, graph_names=None, latex_names=None, metrics_names=None, folder='./'):
        '''dumps global metrics into a fancy latex table'''

        if not graph_names:
            graph_names = list(self.gdict.keys())

        if not latex_names:
            latex_names = graph_names
        elif len(latex_names) != len(graph_names):
            print('warning: graph names and latex names have different length! latex names will be ignored')
            latex_names = graph_names
            

        if not metrics_names:
            print('warning: no metrics names specified! will use metrics dict keys')
            name = graph_names[0]
            try:
                metrics_names = list(self.gdict[name]['global_metrics'].keys())
            except:
                print('warning: global metrics not found for graph {}! computing all metrics...'.format(name))
                self.get_metrics([name])
                metrics_names = list(self.gdict[name]['global_metrics'].keys())

        data = {}
        data['dataset'] = metrics_names
        for i,name in enumerate(latex_names):
            try:
                data[name] = list(self.gdict[graph_names[i]]['global_metrics'].values())
            except:
                print('warning: global metrics not found for graph {}! computing all metrics...'.format(graph_names[i]))
                self.get_metrics([graph_names[i]])
                data[name] = list(self.gdict[graph_names[i]]['global_metrics'].values())

        df = pd.DataFrame(data)
        df.set_index('dataset',inplace=True)
        df = df.transpose()
        strtypes = [0,1,3,4,5,6,14,15,16]
        floattypes = [c for c in range(20) if c not in strtypes]
        dtypes = {df.columns[c]:int for c in strtypes}
        dtypes.update({df.columns[c]:float for c in floattypes})
        df = df.astype(dtypes)
        df = df.round(5)
        df = df.astype('str')
        df = df.transpose()
        tab = df.to_latex(bold_rows=False, column_format='lrrrr', escape=False)
        with open(folder+'global_metrics.tab', 'w') as fout:
            print(tab, file=fout)
            

##################### PLOT #######################

    def set_plot_profile(self, unit=5):
        # small = 4
        # normal = 5
        # large = 6
        plot_params = {}
        plot_params['title_font_size'] = 4*unit # 20
        plot_params['tick_labels_font_size'] = 3.6*unit # 18
        plot_params['labels_font_size'] = 6*unit # 30
        plot_params['legend_font_size'] = 2.8*unit # 14
        plot_params['scatter_point_size'] = 16*unit # 80
        plot_params['annotate_font_size'] = 2.4*unit # 12
        plot_params['labels_padding'] = 6*unit # 30

    def plot_pairwise_centralities(self, graph_names=None, plot_names=None, centralities_names=None, annotate_with='url', folder='./'):

        if not graph_names:
            graph_names = list(self.gdict.keys())

        if not plot_names:
            plot_names = graph_names
        elif len(plot_names) != len(graph_names):
            print('warning: graph names and plot names have different length! plot names will be ignored')
            plot_names = graph_names

        if not centralities_names:
            print('warning: no centralities names specified! will use centralities dict keys')
            name = graph_names[0]
            try:
                centralities_names = list(self.gdict[name]['centralities'].keys())
            except:
                print('warning: centralities metrics not found for graph {}! computing all metrics...'.format(name))
                self.get_metrics([name])
                centralities_names = list(self.gdict[name]['centralities'].keys())

        name = graph_names[0]        
        mcor = list(self.gdict[name]['centralities'].keys())
        num  = len(graph_names)
        rows = 1
        cols = 1
        while rows*cols<num:
            if rows == cols:
                cols += 1
            else:
                rows += 1

        data = [[] for _ in range(num)]  
        for index,name in enumerate(graph_names):
            for j,cen in enumerate(mcor):
                data[index].append(np.array(self.gdict[name]['centralities'][cen]))
                data[index][j] = data[index][j]/max(data[index][j])

        plt.gca().set_aspect('equal') # adjustable='box')
        for j,cen1 in enumerate(mcor):
            for k,cen2 in enumerate(mcor):
                if k>j:
                    # print(cen1, cen2)

                    points = [[] for _ in range(num)]
                    radius = 0.85
                    minvalue = 0.4    
                    for index,name in enumerate(graph_names):
                        for i in range(len(data[index][j])):
                            p = [data[index][j][i],data[index][k][i]]
                            if np.linalg.norm(p)>radius and min(p[0],p[1])>minvalue:
                                points[index].append((p,self.gdict[name]['igraph'].vs[i][annotate_with]))

                    fig = plt.figure(figsize=(12,12))
                    bigax = fig.add_subplot(111)
                    # Turn off axis lines and ticks of the big subplot
                    bigax.spines['top'].set_color('none')
                    bigax.spines['bottom'].set_color('none')
                    bigax.spines['left'].set_color('none')
                    bigax.spines['right'].set_color('none')
                    bigax.tick_params(labelcolor='w', top='off', bottom='off', left='off', right='off')

                    ###
                    ax  = []
                    circle_rad = 8  # This is the radius, in points
                    for index in range(num):
                        pos = index+1
                        ax.append(fig.add_subplot(rows,cols,pos))
                        ax[index].scatter(data[index][j],data[index][k], alpha=0.4, c='b', s=80, marker='+')
                        ax[index].set_title(plot_names[index], fontsize=20)
                        ax[index].tick_params(labelsize=18)
                        for point in points[index]:
                            ax[index].plot(point[0][0], point[0][1], 'o', ms=circle_rad*2, mec='r', mfc='none', mew=1)
                            ax[index].text(point[0][0], point[0][1]-0.1, point[1], backgroundcolor="white",
                                        ha='center', va='top', color='b', fontsize=12)

                    # Set common labels
                    bigax.set_xlabel(centralities_names[j], labelpad=30, fontsize=30)
                    bigax.set_ylabel(centralities_names[k], labelpad=30, fontsize=30)
                    fig.savefig(folder+cen1+'_'+cen2+'.png', bbox_inches='tight') #, dpi=150)
                    # print()
                    
                    
    def plot_centrality_distribution(self, graph_names=None, plot_names=None, centralities=None, centralities_names=None, fibonacci_binning=False, powerlaw_fit=False, folder='./'):

        if not graph_names:
            graph_names = list(self.gdict.keys())

        if not plot_names:
            plot_names = graph_names
        elif len(plot_names) != len(graph_names):
            print('warning: graph names and plot names have different length! plot names will be ignored')
            plot_names = graph_names

        if not centralities:
            name = graph_names[0]        
            try:
                centralities = list(self.gdict[name]['centralities'].keys())
            except:
                print('warning: centralities metrics not found for graph {}! computing all metrics...'.format(name))
                self.get_metrics([name])
                centralities = list(self.gdict[name]['centralities'].keys())

        if not centralities_names:
            print('warning: no centralities names specified! will use centralities dict keys')
            name = graph_names[0]
            try:
                centralities_names = list(self.gdict[name]['centralities'].keys())
            except:
                print('warning: centralities metrics not found for graph {}! computing all metrics...'.format(name))
                self.get_metrics([name])
                centralities_names = list(self.gdict[name]['centralities'].keys())

        if fibonacci_binning is True:
            fibonacci_binning = centralities
        elif not fibonacci_binning:
            fibonacci_binning = []
        
        if powerlaw_fit is True:
            powerlaw_fit = centralities
        elif not powerlaw_fit:
            powerlaw_fit = []
  
        ### set plot grid parameters ###
        num  = len(graph_names)
        rows = 1
        cols = 1
        while rows*cols<num:
            if rows == cols:
                cols += 1
            else:
                rows += 1
                                   
        #### data extraction and plots ####
        plt.gca().set_aspect('equal')
        cen_index = 0
        for cen in centralities:
            fig = plt.figure(figsize=(12,12))
            bigax = fig.add_subplot(111)
            # Turn off axis lines and ticks of the big subplot
            bigax.spines['top'].set_color('none')
            bigax.spines['bottom'].set_color('none')
            bigax.spines['left'].set_color('none')
            bigax.spines['right'].set_color('none')
            bigax.tick_params(labelcolor='w', top='off', bottom='off', left='off', right='off')

            ax  = []
                   
            for index,name in enumerate(graph_names):
                
                pos = index+1
                ax.append(fig.add_subplot(rows,cols,pos))
                ax[index].set_xscale('log')
                ax[index].set_yscale('log')
                ax[index].set_title(plot_names[index], fontsize=20)
                                
                #### scatter plot of data distribution ####
                d = np.array(self.gdict[name]['centralities'][cen])
                xdata, ydata = get_dist(d)
                if xdata[0]==0:
                    prob0 = '={:.3f})'.format(ydata[0])
                else:
                    prob0 = '=0)'
                ax[index].scatter(xdata, ydata, s=80, c='b', marker='+', alpha=0.4, label='Data (' + r'$\Pr[0]$' + prob0)
                # ylimbottom = ax[index].get_ylim()[0]
                ymindata = min(ydata)
                
                #### scatter plot of fibonacci binning ####
                if cen in fibonacci_binning:                    
                    if issubclass(d.dtype.type, np.integer):
                        maxbin = max(d)
                    else:
                        maxbin = int(len(d)/10)
                    xfib, yfib = fibonacci_hist_cont(d,maxbin)
                    ax[index].scatter(xfib, yfib, s=80, c='g', marker='o', label='Fib. binning') 
                    yminfib = min([y for y in yfib if y>0])
                    # ylimbottom = min(ymindata,yminfib)-(ymindata-ylimbottom)

                
                #### plot of powerlaw fit ####
                if cen in powerlaw_fit:
                    if issubclass(d.dtype.type, np.integer):
                        discrete = True
                    else:
                        discrete = False
                    fit = fit_data(d, discrete=discrete)
                    alpha = fit.alpha
                    xmin  = fit.xmin
                    xmax = xdata[-1]
                    scale = sum(ydata[xdata.index(xmin):xdata.index(xmax)])
                    xfit, yfit = get_fit_dist(alpha, xmin, xmax, scale) 
                    ##
                    if issubclass(d.dtype.type, np.floating):
                        rebeta = (alpha-1)/(xmin**(1-alpha)-xmax**(1-alpha)) * scale
                        rescale = 0
                        fakecount = 0
                        for fakei in range(len(xfib)):    
                            if yfib[fakei] > 0:
                                xfake = xfib[fakei]            
                                yfake = rebeta*xfake**(-alpha)           
                                rescale += yfake/yfib[fakei]
                                fakecount += 1
                        rescale /= fakecount
                        yfit /= rescale
                    if cen in fibonacci_binning:
                        ymindata = min(ymindata, yminfib)
                    # print(len(yfit), min(yfit), ymindata)
                    # print(yfit)
                    xfit = xfit[yfit>=ymindata]
                    yfit = yfit[yfit>=ymindata]
                    # print(len(yfit), min(yfit), ymindata)
                    ax[index].plot(xfit, yfit, 'm--', label='PL fit (' + r'$\alpha$' +'={:.2f})'.format(alpha))
                # ax[index].set_ybound(lower=ylimbottom)
                # print(cen, ylimbottom)
                ax[index].legend(loc='upper right', prop={'size': 14})
                ax[index].tick_params(labelsize=18)

            # Set common labels
            bigax.set_xlabel(centralities_names[cen_index], labelpad=30, fontsize=30)
            cen_index += 1
            bigax.set_ylabel('Probability', labelpad=40, fontsize=30)
            fig.savefig(folder+cen+'.png', bbox_inches='tight') #, dpi=150)
                    
                    
    def plot_pie_diagram(self, select_by='url', plot_names=None, seed=250, folder='./'):
        
        def powerset(iterable):
            "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
            s = list(iterable)
            return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))
        
        def intensity(t,jump=6):
            i = 0
            c = 1
            while(len(t)>i+1):
                diff = t[i+1]-t[i]-1
                if diff==0:
                    c += 1
                else:
                    c += jump*diff
                i += 1
            return c
    
        graph_names = [name for name in self.gdict.keys() if name!='intersection']
        if not plot_names:
            plot_names = graph_names
        elif len(plot_names) != len(graph_names):
            print('warning: graph names and plot names have different length! plot names will be ignored')
            plot_names = graph_names

        graphs = [self.gdict[name]['igraph'] for name in self.gdict if name!='intersection']

        values = []
        for graph in graphs:
            values.append(set(graph.vs[select_by]))

        indices = [t for t in powerset(range(len(graphs))) if len(t)>0]
        intersections = {}
        for t in indices:
            intersections[t] = set.intersection(*[values[u] for u in t])
            if t != indices[-1]:
                intersections[t] = set.difference(intersections[t],set.union(*[values[u] for u in indices[-1] if u not in t]))

        lens = [len(v) for v in intersections.values()]
        data = [(k,lens[i],'{:.2f}'.format(100*np.round(lens[i]/sum(lens),4))) for i,k in enumerate(intersections.keys())]

        # workaround for reordering pie slices
        r = random.Random()
        r.seed(seed)
        r.shuffle(data)

        #####
        fig, ax = plt.subplots(figsize=(14,11))
        cmap = plt.get_cmap("Reds")
        cunit = 30
        colors = [cmap(cunit*intensity(d[0])) for d in data] 
        cblank = 'w' # 'black'

        size = 1/len(plot_names)

        vals = [d[1] for d in data] 

        patches = []
        for index in range(len(plot_names)):
            cols = [colors[i] if index in d[0] else cblank for i,d in enumerate(data)]
            patches.append(ax.pie(vals, radius=1-index*size, colors=cols, startangle=90, wedgeprops=dict(edgecolor='w', width=0.31))[0])

        notes = [' and '.join(plot_names[i] for i in d[0])+(len(d[0])<len(plot_names)-1)*' only'+'\n'+r'${}\%\;({}\;hs)$'.format(d[2],d[1]) for d in data] 

        bbox_props = dict(boxstyle="square,pad=0.6", fc="w", ec="k", lw=0.72)
        kw = dict(xycoords='data', textcoords='data', arrowprops=dict(arrowstyle="-"),
              bbox=bbox_props, zorder=0, va="center")

        ######### the following part may require manual tuning for different number of graphs ########
        patches = patches[-1]
        pang = 0
        for i, p in enumerate(patches):
            ang = (p.theta2 - p.theta1)/2. + p.theta1
            y = np.sin(np.deg2rad(ang))
            x = np.cos(np.deg2rad(ang))
            horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
            if abs(ang-pang) < 5:
                diff = 0.2 
            elif i == 0:
                diff = -0.2
            else:
                diff = 0
            connectionstyle = "angle,angleA=0,angleB={}".format(ang)
            kw["arrowprops"].update({"connectionstyle": connectionstyle})
            ax.annotate(notes[i], xy=(x, y), xytext=((1.1+diff)*(x/2+np.sign(x)/2), (1.3-diff)*y), horizontalalignment=horizontalalignment, fontsize=15, **kw)
            pang = ang

        ax.set(aspect="equal")
        labels = 'seen once', 'partially persistent', 'fully persistent', 'reappeared'
        order = [2,1,5,0]
        ax.legend([patches[i] for i in order], labels, fontsize=15, loc=(0.31,-0.05))
        fig.savefig(folder+'pie_chart.png') #, dpi=150)
        
        
    def plot_cumulative_top_out_degree(self, n_top=25, graph_names=None, plot_names=None, folder='./'):
        
        if not graph_names:
            graph_names = list(self.gdict.keys())
        
        num  = len(graph_names)
        rows = 1
        cols = 1
        while rows*cols<num:
            if rows == cols:
                cols += 1
            else:
                rows += 1
        fig = plt.figure(figsize=(12,12))
        bigax = fig.add_subplot(111)
        # Turn off axis lines and ticks of the big subplot
        bigax.spines['top'].set_color('none')
        bigax.spines['bottom'].set_color('none')
        bigax.spines['left'].set_color('none')
        bigax.spines['right'].set_color('none')
        bigax.tick_params(labelcolor='w', top='off', bottom='off', left='off', right='off')
        ax  = []

        for index,name in enumerate(graph_names):
            pos = index+1
            ax.append(fig.add_subplot(rows,cols,pos))
            ax[index].set_title(plot_names[index], fontsize=20)
            
            g = self.gdict[name]['igraph']
            num_nodes = g.vcount()
            outdegree = g.degree(mode=OUT)
            g_outdeg = pd.DataFrame({'out_degree':outdegree, 'id':range(num_nodes)})
            g_outdeg = g_outdeg.sort_values(by=['out_degree'], ascending=False)
            g_outdeg = g_outdeg.reset_index(drop=True)

            neighbors = set()
            #n_nodes = []
            n_nodes_perc = []
            for x in range(num_nodes):
                node = g_outdeg['id'][x]
                for z in self.gdict[name]['igraph'].vs[node].neighbors(mode=OUT):
                    neighbors.add(z.index) 
                n_nodes_perc.append(len(neighbors)/num_nodes*100 )
            
            x = [ z+1 for z in range((n_top-1)) ]
            ax[index].plot( x, n_nodes_perc[:(n_top-1)], 'o--')
            ax[index].set_ylim( (min(n_nodes_perc[:(n_top-1)])-5), 100)
            ax[index].tick_params(labelsize=18)
            
        bigax.set_ylabel("Nodes linked %", labelpad=30, fontsize=30)
        bigax.set_xlabel("Number of out-hubs", labelpad=30, fontsize=30)
        fig.savefig(folder+'cumulative_top.png', bbox_inches='tight') #, dpi=150)
        plt.close()
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        