import sys
from collections import Counter
import itertools

import numpy as np
from igraph import *
from tabulate import tabulate
import configparser

import powerlaw
import matplotlib.pyplot as plt

import pickle

# TOREROS MACRO

TOREROS_EDGES_ATTR  = ['label']
TOREROS_NODES_ATTR  = ['nodeId', 'url', 'code', 'record', 'href', 'src', 'data', 'http', 'location']
TOREROS_NODES_STAT  = ['record', 'langTika', 'langDet', 'wordCount']
TOREROS_KEY         = 'nodeId'
TOREROS_VECTOR_CAT  = ['Social Network', 'Information System', 'Fraud', 'Gambling', 'Murder and grievous bodily injury', 'Sexual exploitation of children and child pornography', 'Illicit trafficking in weapons and munitions and explosives', 'Rape', 'Racism and xenophobia', 'IT Services companies', 'Media companies', 'Terrorism', 'Weaponary', 'Illicit drugs', 'Cyber Attack', 'Cyber deception', 'Cyber Security'] 
TOREROS_VECTOR_KEY = 'record'

# attribute - Nodes Class
class aNodes:
    '''
        Class for a graph in which each node/edge has a set of attributes.
        Possibly, nodes are associated with a (semantic) vector.  An aNodes object
        can be built starting from either a list of nodes, vectors or edges. When
        vectors or edges are used, nodedict is automatically initialized with empty
        attributes.

        Attributes:
        - nodedict is a nested dictionary whose items are the vertices. NB: keys must
        be integers!
        - nodekey is the key used to index nodedict. Can be a number or one of the
        attributes.
        - node_attributes is the list of possible attributes
        - vectordict is a dictionary with same keys as nodedict and whose values are
        (semantic) vectors
        - vector_categories is the list of categories corresponding to the
        coordinates of the vectors in vectordict
        - edges is the list of edges
        - edges_attributes is the list of possible attributes for edges
        - stats is a dictionary for graph statistics
    '''    

    def __init__(self, 
                 nodedict   = {}, node_attributes   = [], node_key    = '',   
                 vectordict = {}, vector_categories = [], 
                 edgelist   = [], edge_attributes   = [], 
                 statdict   = {}
                 ):
        self.nodedict          = nodedict
        self.node_attributes   = node_attributes[:]
        self.node_key          = node_key
        self.vectordict        = vectordict
        self.vector_categories = vector_categories[:]
        self.edges             = edgelist
        self.edge_attributes   = edge_attributes[:]
        self.stats             = statdict

    def __repr__(self):
        nnodes = 'number of nodes: ' + str(len(self.nodedict)) + '\n'
        nedges = 'number of edges: ' + str(len(self.edges))    + '\n'  
        key    = 'data key:        ' + str(self.node_key)      + '\n'
        natt   = 'nodes attributes:' + str(self.node_attributes) + '\n'
        eatt   = 'edges attributes:' + str(self.edge_attributes) + '\n'

        header = '\n' + self.__class__.__name__ + ' : ' + str(aNodes) + '\n'

        all  = header + nnodes + nedges + key + natt + eatt

        return all

    
    def summary(self):
        return self.__repr__()
        


    @classmethod
    def getNodes(cls, node_file = str, attributes = [], dkey = ''):
        """ read nodes from file using the read_nodes function and returns an aNodes object.
            (alternative to get getVectors and getEdges)            

            To add more attributes to nodedict from another file, use the
            method: addAttributeFromFile.
        """
        anode = cls()
        anode.node_attributes = attributes[:]
        anode.nodedict, anode.node_key = anode.read_nodes(f_name = node_file, attributes = anode.node_attributes, nodedict = {}, dkey = dkey)
        return anode


    @classmethod
    def getVectors(cls, vectors_file = str, vector_categories = [], categories_from_file = False):
        """ read vectors from file using the read_vectors function and returns an aNodes object
            (alternative to getNodes and getEdges)

            The nodedict is initialized with all the keys found in the vector file.
            (See the documentation of read_vectors for info on the vector file).

            If categories_from_file is True then the categories are read from the first
            line of the file as a list.

            vector_categories is used to initialize anode.vector_categories, if is empty
            then if categories_from_file is True the categories are read from file,
            otherwise the column indices are used.
        """
        anode = cls()

        # Read the vectors from file 
        anode.vectordict, anode.vector_categories = anode.read_vectors(f_name = vectors_file, categories_from_file = categories_from_file)

        if bool(vector_categories):
            anode.vector_categories = vector_categories

        # Create nodes
        for k in anode.vectordict.keys():
            anode.nodedict.setdefault(k, {})

        return anode


    @classmethod
    def getEdges(cls, edges_file = str, edge_attributes = [], node_key = '', compute_edge_multeplicity = False):
        """ read edges from file using the read_edges function and returns an aNodes object
            (alternative to getNodes and getVectors)
            
            Each edge is stored with attributes in the edge dictionary of the aNode struct.
            Each node is stored in the nodedict dictionary and three attributes are created for
            each node:
            ingoing degree, outgoing degree, all degree.
            This method uses the functions read_edges and nodes_degrees_from_edges to fill the
            data struct, see the respective documentations form more infos.

            If compute_edge_multeplicity is True return a dictionary with edge multeplicity
            (see the doc of nodes_degrees_from_edges)

            If you already have the aNode data struct use the alternative function addEdgesFromFile
            and/or directly the function nodes_degrees_from_edges (this latter case imply that
            you have already read the edges into a list).

            Frankly the most common situation is that you first use this function to read edges
            and compute degrees and then you use the function addAttributeFromFile to add nodes
            attributes.
        """
        anode = cls()
        if bool(edge_attributes):
            if isinstance(edge_attributes,str):
                anode.edge_attributes = edge_attributes.split()
            else:
                anode.edge_attributes = edge_attributes[:]
        else:
            anode.edge_attributes = []
           # anode.edge_attributes = None
        anode.node_key = node_key       
        anode.edges    = anode.read_edges(edges_file, anode.edge_attributes)
        anode.nodedict = anode.nodes_degrees_from_edges(anode.edges, nodes = {}, edge_mult = compute_edge_multeplicity)
        # print(anode.nodedict[0])
        anode.node_attributes += ['outdeg', 'indeg', 'alldeg', 'outdegsimply', 'indegsimply']
        # print(anode.nodedict[0])

        return anode

    
    def addAttributesFromFile(self, nodes_file = str, attributes = [], dkey = '', oldkey = ''):
        """ update nodedict of an existing aNode object adding nodes and/or attributes from file"""
        attributes = attributes[:]
        print("oldkey = ", self.node_key)
        self.nodedict, self.node_key = self.read_nodes(f_name = nodes_file, attributes = attributes, nodedict = self.nodedict, dkey = dkey, oldkey = self.node_key)
        #try:
        #    attributes.remove(self.node_key)
        #except:
        #    pass

        if self.node_attributes:
            for att in attributes:
                self.node_attributes.append(att)
        else:
            self.node_attributes = attributes

        # return self

    
    def addVectorsFromFile(self, vectors_file, vectors_key='', vector_categories=[], categories_from_file=False, normalize=False):
        """ loads the vectordict of an existing aNode object from file"""
        remap = False
        if not vectors_key:
            print("addVectorsFromFile: WARNING! No key specified for vectors! Currently in-use key will be used")        
        elif self.node_key != vectors_key:
            remap = True

        if isinstance(vector_categories, str):
            vector_categories = [cat.strip('[]\n ').lower() for cat in vector_categories.split(',')]
        
        vdict, self.vector_categories = self.read_vectors(f_name=vectors_file, vector_categories=vector_categories, categories_from_file=categories_from_file)
        
        if remap:
            self.vectordict = {k:vdict.get(v.get(vectors_key)) for k,v in self.nodedict.items()}
        else:
            self.vectordict = {k:vdict.get(k) for k in self.nodedict}  

        if normalize:
            max_vec = 0
            for v in self.vectordict.values():
                if v is not None:
                    max_v = max(v)
                if max_v > max_vec:
                    max_vec = max_v
            if max_vec > 0:
                for v in self.vectordict.values():
                    if v is not None:
                        v = [i/max_vec for i in v]
        

    def iGraphObject(self, load_vectors=False):
        """ This method create an iGraph object using the function create_igraph_object.
            Please refer directly to the function documentation.
        """
        if load_vectors:
            self.node_attributes.append('vector')
            for n in self.nodedict:
                self.nodedict[n]['vector'] = self.vectordict[n]

        g = self.create_igraph_object(self.nodedict, self.edges, self.node_attributes, self.node_key, self.edge_attributes)
        
        if load_vectors:
            g['legend'] = self.vector_categories
            l = len(g.vs.find(vector_ne=None)['vector'])
            g['mean vector'] = [np.mean([a[i] for a in g.vs['vector'] if a is not None]) for i in range(l)]
            g['std vector'] = [np.std([a[i] for a in g.vs['vector'] if a is not None]) for i in range(l)]         
            
        return g

    
    def removeVertices(self, vset):
        ''' remove all vertices in vset and the corresponding edges '''
        
        for v in vset:
            del(self.nodedict[v])
        vset = set(vset)
        self.edges = [e for e in self.edges if e[0] not in vset and e[1] not in vset]
        
    
#    def getLayers(self):
#        ''' split the graph into layers based on edges attributes '''            
#       
#        degrees = ['outdeg', 'indeg', 'alldeg', 'outdegsimply', 'indegsimply']
#        attributes = [i for i in self.node_attributes if i not in degrees]
#        layers = {att:aNodes(node_attributes=attributes, node_key=self.node_key, vector_categories=self.vector_categories) for att in self.edge_attributes}
#        for e in self.edges:
#            u, v, types = e[0], e[1], e[2]
#            for t in types:
#                layers[t].nodedict[u] = self.nodedict[u] 
#                layers[t].nodedict[v] = self.nodedict[v]
#
#            layers[att].edges.append((e[0],e[1])
#        
#set degrees to zero and call get_degrees
        
                    


    def getSubgraphByAttribute(self, attribute, value):
        """[BETA]
           
        extract the subgraph induced by the set of vertices whose specified attribute is in the specified (range of) values"""
        subnodedict = dict()
        if not isinstance(attribute, list):
            value = [value]
        for k,v in self.nodedict.items():
            if attribute in v and v[attribute] in value:
                subnodedict[k] = v
        s = aNodes()
        s.nodedict = subnodedict
        s.node_attributes = self.node_attributes[:]
        s.node_key = self.node_key
        s.vectordict = {k:v for k,v in self.vectordict.items() if k in s.nodedict}
        s.vector_categories = self.vector_categories[:]
        s.edges = [edge for edge in self.edges if (edge[0] in s.nodedict and edge[1] in s.nodedict)]
        s.edge_attributes = self.edge_attributes[:]
        s.stats = dict()
        return s


    @staticmethod
    def read_nodes(f_name, attributes = [], dkey = None, nodedict = {}, oldkey = None):
        """ 
            read a list of nodes with attributes from file, creating or extending a nodedict.

            Args:
                f_name:         (str) input filename
                attributes:     (list) a list with the name of each field to read, if empty indices
                                are used as key for the nested dictionary
                dkey:           the field in attributes used as the key of the nodes dictionary
                                MUST be one of the attributes or None
                                if dkey is None or is not in attributes the first attribute is used 
                                when creating a new dictionary, whereas the old key is used when 
                                extending an already existing one
                nodedict:       (dict) the input/output dictionary, if not empty it is extended by 
                                adding new fields to the existing nodes or new nodes
                oldkey:         the key of the input nodedict, mandatory if nodedict is not empty

            FILE format:
                Each line must have the same number of columns 
                each column specifies the value for a different attribute
                if the number of columns and len(attributes) are different, the minimum is used            
                values of column dkey are treated as integers (this may raise exceptions)

            Return:
                A dictionary nodes = {dkey1:{field1, field2, field3}, dkey2:{field1, field2, field3}, ...}
        """
        
        #print("read_nodes: reading nodes form file {} ...".format(f_name))
        if nodedict:
            if type(nodedict) is dict:
                pass
            else:
                print("read_nodes: the input data must be a dictionary! Quit")
                sys.exit()

        with open(f_name) as f:
            ncols = len(f.readline().split())
            #print("read_nodes: there are {} columns".format(ncols))

        if not attributes:
            print("read_nodes: WARNING, no attributes! Fallback to indices")
            attributes = [i for i in range(ncols)]
        elif ncols != len(attributes):
            print("read_nodes: WARNING,  The number of attributes differ from the number of columns, extra attributes or extra columns will be ignored")
            if ncols > len(attributes):
                ncols = len(attributes)

        attributes = attributes[:ncols]
                  
        if nodedict:
            if not oldkey:
                print(nodedict)
                print("read_nodes: ERROR! No key specified for the existing nodedict! Quit.")
                sys.exit()
            oldattributes = set(next(iter(nodedict.values())).keys())
            if not dkey or dkey not in attributes:
                print("read_nodes: WARNING! Specified key is None or not in attributes, keeping in-use key")
                newkey = oldkey
                if oldkey not in attributes:
                    map_key = None
                    for att in oldattributes:
                        if att in attributes:
                            map_key = att
                            break
                    if not map_key:
                        print("read_nodes: ERROR! No common attribute found! Merge is impossible! Quit.")
                        sys.exit()
                    else:
                        print("read_nodes: Updating nodedict with new attributes. New nodes will not be added because in-use key '{}' is not in attributes. Mapping key is '{}'.".format(oldkey,map_key))
                        map_index = attributes.index(map_key)
                        attributes.remove(map_key)
                        new_dict = {}
                        with open(f_name) as f:
                            for line in f:
                                words = [int(w) if w.isdigit() else w for w in line.split()]
                                k = words.pop(map_index)
                                new_dict.setdefault(k,{})
                                for i in range(len(attributes)):
                                    new_dict[k][attributes[i]] = words[i]
                        for node_atts in nodedict.values():
                            node_atts.update(new_dict.get(node_atts[map_key],{}))
                else:
                    print("read_nodes: Updating nodedict with new attributes and/or new nodes. Keeping in-use key '{}'.".format(oldkey))
                    k_index = attributes.index(oldkey)
                    attributes.remove(oldkey)
                    with open(f_name) as f:
                        for line in f:
                            words = [int(w) if w.isdigit() else w for w in line.split()]
                            k = int(words.pop(k_index))
                            nodedict.setdefault(k,{})
                            for i in range(len(attributes)):
                                nodedict[k][attributes[i]] = words[i]
            else:
                if dkey == oldkey:
                    newkey = oldkey
                    print("read_nodes: Updating nodedict with new attributes and/or new nodes. New key is equal to already in-use key '{}'.".format(oldkey))
                    k_index = attributes.index(oldkey)
                    attributes.remove(oldkey)
                    print(attributes)
                    with open(f_name) as f:
                        for line in f:
                            words = [int(w) if w.isdigit() else w for w in line.split()]
                            k = int(words.pop(k_index))
                            nodedict.setdefault(k,{})
                            for i in range(len(attributes)):
                                nodedict[k][attributes[i]] = words[i]
                else: 
                    newkey = dkey
                    k_index = attributes.index(dkey)
                    attributes.remove(dkey)
                    new_dict = {}
                    with open(f_name) as f:
                        for line in f:
                            words = [int(w) if w.isdigit() else w for w in line.split()]
                            k = int(words.pop(k_index))
                            new_dict.setdefault(k,{})
                            for i in range(len(attributes)):
                                new_dict[k][attributes[i]] = words[i]
                    if dkey not in oldattributes:
                        print("read_nodes: WARNING! Specified key '{}' differs from in-use key '{}' and is not among in-use attributes! Only new nodes will be kept, but old attributes will be inherited when possible.".format(dkey,oldkey))
                        map_key = None
                        for att in attributes:
                            if att in oldattributes:
                                map_key = att
                                break
                        if not map_key:
                            print("read_nodes: WARNING! No common attribute found! Cannot inherit any attribute. New nodedict will simply substitute old one.")
                    else:
                        map_key = dkey    
                    if map_key:
                        for att in attributes:
                            try:
                                oldattributes.remove(att)
                            except:
                                pass
                        new_nodedict = {}
                        for node, node_atts in nodedict.items():
                            k = int(node_atts[map_key])
                            v = {oldkey:node}
                            v.update({att:node_atts.get(att) for att in oldattributes})
                            new_nodedict[k] = v
                        for node, node_atts in new_dict.items():
                            node_atts.update(new_nodedict.get(node,{}))
                    nodedict = new_dict
        else:
            if not dkey or dkey not in attributes:
                print("read_nodes: WARNING! No key specified or key not in attributes! Key is set to first attribute")
                dkey = attributes[0]
                attributes = attributes[1:]
                k_index = 0
            else:
                k_index = attributes.index(dkey)
                attributes.remove(dkey)
            with open(f_name) as f:
                for line in f:
                    words = [int(w) if w.isdigit() else w for w in line.split()]
                    k = int(words.pop(k_index))
                    nodedict.setdefault(k,{})
                    for i in range(len(attributes)):
                        nodedict[k][attributes[i]] = words[i]
            newkey = dkey

        return nodedict, newkey


    @staticmethod
    def read_edges(filename, attributes = []):
        """ read an edge list with attributes from the input file.
            Each line of the input file has the form:

            u v attr1 attr2 ...

            where u and v are integers corresponding to vertices and attrX is
            an attribute associated with the edge (u, v). Atrributes are read
            as strings.

            each line is treated as a new edge even when (u,v) already exists 
            thus multigraphs are admitted.

            Note that all lines must have the same number of columns (at least 2).

            Args:
                filename: (str) the name of the input file
                attributes: (list) a list of the field to read, beyond the edge.

            Return:
                A list of 3-tuple:

                [(u,v,[attr1, attr2, ..]), (z,x,[attr1, attr2, ...])

        """


        expected_cols = 2 + len(attributes)
        with open(filename) as f:
            line = f.readline()
            data = line.split()
            ncols = len(data)
            mincols = min(expected_cols,ncols)
            if ncols<2:
                print("read_edges: ERROR, the input file must have at least two columns! Quit.", file = sys.stderr())
                sys.exit()
            elif ncols!=expected_cols:
                print("read_edges: WARNING, expecting a file with {} columns, but the columns are {}.\n"
                "Read only first {} columns".format(expected_cols, ncols, mincols))

        edges = []
        with open(filename) as f:
            for line in f:
                data = line.split()
                u = int(data[0])
                v = int(data[1])
                info = data[2:]

                if mincols == 2:
                    edge = (u, v)
                elif mincols == 3:
                    edge = (u, v, info[0])
                else:
                    a = []
                    for k in range(2, mincols):
                        a.append(data[k])
                    edge = (u,v,a)
                edges.append(edge)

        print("read_edges: number of edges read from file {}: {}".format(filename, len(edges)))
        return edges

    @staticmethod
    def nodes_degrees_from_edges(edges, nodes = {}, edge_mult = False):
        """ read a list of tuples of the form (u, v, [attr1, attr2, ...]),
            where each tuple represents an edge with its attributes, and computes the degree
            of each node.
            Five degrees are computed for each node:
            outdeg:         number of outcoming edges (counting multiplicity)
            indeg:          number of incoming edges  (counting multiplicity)
            alldeg:         number of edges in the undirected graph (no multiplicity)
            outdegsimply:   number of outcoming edges (no multiplicity)
            indegsimply:    number of outcoming edges (no multiplicity)

            This function is meant to be used along with read_edges that produces the tuples.

            Args:
                edges:  (list) a list of 3-tuple (u (str/int), v(str/int), [attr1(str), attr2, ...]) 
                nodes:  (dict) a nestedd dictionary of nodes with attributes. If not empty nodes
                        are updated with the new fields and new nodes are added.
                        nodes={dkey1:{attr1: , attr2: , ...}, dkey2:{attr1:, attr2:, ...},  ...}
                edge_mult: (bool) if True compute the multeplicity of edges, see below

            Return:
                nodes: the updated nodes dictionary.
                [if edge_mult is True]
                newedges: a dictionary that stores the edge multiplicity. The keys of the 
                          dictionary are the edge tuples: (u,v).

                          newedges = {(u,v):{out: , in:, all: } }

                          out stores the multeplicity of the edge (u, v)
                          in  stores the multeplicity of the edge (v, u)
                          all stores the multeplicity of (u, v) in the 'undirected' graph.
           
           Note: actually to compute the degrees correctly I need to store also the edge
                 count so the edge multeplicity must be computed anyway.
        """

        newedges = dict()   

        # Note:
        # To ensure that all the new fields are added to exisisting node one can
        # process the nodes key with the following code line but it turns out that
        # this method is a little bit slower than update during the read. 
        # for k in nodes:
        #    nodes[k].update(outdeg=0, indeg=0, alldeg=0, outdegsimply=0, indegsimply=0)

        for e in edges:
            u = int(e[0])
            v = int(e[1])

            ee = (u, v)
            re = (v, u)


            # If the edge (u,v) is already in the edge list, update the values of
            # outdeg for u and indeg for v
            if ee in newedges:
                newedges[ee]['out'] += 1
                newedges[ee]['all'] += 1
                nodes[u]['outdeg']  += 1
                nodes[v]['indeg']   += 1
            # If (v, u) is already in the list, update the values of indeg for u
            # and outdeg for v but now u is v!!
            elif re in newedges:
                newedges[re]['in']  += 1
                newedges[re]['all'] += 1
                nodes[u]['outdeg']  += 1
                nodes[v]['indeg']   += 1
                if nodes[u]['outdegsimply'] == 0:
                    nodes[u]['outdegsimply'] = 1
                if nodes[v]['indegsimply'] == 0:
                    nodes[v]['indegsimply']  = 1
            
            # First time we see this edge.
            else:
                # Initialize the edge dictionary
                newedges.setdefault(ee, {'out':1, 'in':0, 'all':1})

                # If the node is already in the node dictionary update it's attributes
                if u in nodes:
                    if 'outdeg' not in nodes[u].keys():
                        nodes[u].update(outdeg=0, indeg=0, alldeg=0, outdegsimply=0, indegsimply=0)
                # First time we see this node, create it with all the attributes. Here we add
                # also nodeId, this will be used in the read_node method
                else:
                    nodes.setdefault(u, {'outdeg':0, 'indeg':0, 'alldeg':0, 'outdegsimply': 0, 'indegsimply': 0}) #, 'nodeId':u})

                if v in nodes:
                    if 'outdeg' not in nodes[v].keys():
                        nodes[v].update(outdeg=0, indeg=0, alldeg=0, outdegsimply=0, indegsimply=0)
                else:
                    nodes.setdefault(v, {'outdeg':0, 'indeg':0, 'alldeg':0, 'outdegsimply': 0, 'indegsimply': 0}) #, 'nodeId':v})

                # This is a new edge, add degrees to u and v
                nodes[u]['outdeg']       += 1
                nodes[u]['outdegsimply'] += 1
                nodes[u]['alldeg']       += 1

                nodes[v]['indeg']        += 1
                nodes[v]['indegsimply']  += 1
                nodes[v]['alldeg']       += 1

        print("nodes_degrees_from_edges: number of edges, nodes read: {}, {}".format(len(edges), len(nodes)))

        if edge_mult:
            return nodes, newedges
        else:
            return nodes

        return


    @staticmethod
    def read_vectors(f_name, vector_categories=[], categories_from_file=False):
        """ read line by line the vector file i.e. a file
            of the form:
            key attribute1 attribute2 attribute3 ....
            
            where key is the label/index of a node in the graph and
            attributes are a list of (semantic) attributes associated
            with the node.

            If categories_from_file is False then the function returns
            a list of indices whose length is equal to the number of columns.

            Return:
            	vector dictionary
            	list of categories 
        """
        vectors = dict()

                    
        with open(f_name) as f:
            header = f.readline()
            hasheader = not header.isnumeric()
            if hasheader:
                filecats = [cat.strip('[]\n ').lower() for cat in header.split(',')]
                if not categories_from_file or (vector_categories and vector_categories!=filecats):
                    print("read_vectors: WARNING! Header found! Categories in file header will be used") 
                categories = filecats    
            else:
                nfilecats = len(header.split(maxsplit=1)[1].split(','))
                if categories_from_file:
                    print("read_vectors: WARNING! Cannot read categories from file!")
                if vector_categories:
                    print("read_vectors: Using speficied categories")
                    categories = vector_categories
                    if len(categories) < nfilecats:
                        print("read_vectors: WARNING! Specified categories are fewer than those on file! Extra columns will be ignored")
                    elif len(categories) > nfilecats:
                        print("read_vectors: WARNING! Vectors are shorter than expected! Extra categories will be ignored")
                        categories = categories[:nfilecats] 
                else:
                    print("read_vectors: Fallback to indices")
                    categories = list(range(nfilecats))
                f.seek(0)
            ncats = len(categories)
            #print("read_vectors: there are {} columns".format(ncats))

            for line in f:
                line = line.split(maxsplit=1)
                key = int(line[0].strip(','))
                vectors[key] = eval(line[1])[:ncats]
    
        return vectors, categories    
   
 
    @staticmethod
    def fit_data(data, xmin=None, xmax=None, compare=False, plot=False, output_file=''):
        ''' fit data to a power law
            
            Args:
                data            is the distribution to fit, can be either a list or a dictionary
                                if a list, is interpreted as the list of all outcomes (e.g., [0,1,2,1,1] means pr(0)=pr(2)=0.2, pr(1)=0.6)
                                if a dictionary, items are interpreted as outcome:number_of_occurrences (e.g., {0:1, 1:3, 2:1})
                xmin
                xmax
                compare         another distribution to which the power law fit can be compared; if True or invalid, exponential is used 
                plot            specifies whether the fit must be plotted (both pdf fit and ccdf fit are plotted)
                output_file     the output file for the plot; if empty a default name is used
        '''
        if isinstance(data,dict):
            d = []
            for k,v in data.items():
                d += [k]*v
            data = np.array(d)
        else:
            data = np.array(data)
        if min(data) == 0:
            data += 1
            print('rescaled data+1 to avoid x=0')
        fit = powerlaw.Fit(data=data, discrete=True, xmin=xmin, xmax=xmax)
        print('Power-law best fit to data:')
        print('\t alpha = {}'.format(fit.power_law.alpha))
        print('\t sigma = {}'.format(fit.power_law.sigma))
        if xmin is None:
            print('\t xmin = {}'.format(fit.power_law.xmin))
        if xmax is None:
            print('\t xmax = {}'.format(fit.power_law.xmax))
        if compare:
            if compare == True:
                R, p = fit.distribution_compare('power_law', 'exponential', normalized_ratio = True)
            else:
                try:
                    R, p = fit.distribution_compare('power_law', compare, normalized_ratio=True)
                except:
                    print('the argument to option "compare" is not a valid distribution, will use exponential')
                    R, p = fit.distribution_compare('power_law', 'exponential', normalized_ratio = True)
            print('Normalized loglikelihood ratio: {}'.format(R))
            print('Significance value: {}'.format(p))
        if plot:
            fig, ax = plt.subplots(nrows=1, ncols=1,figsize=(10,10))
            fit.plot_pdf(color= 'b', ax=ax, label='data pdf')
            fit.power_law.plot_pdf(color= 'b',linestyle='--',label='fit pdf', ax=ax)
            fit.plot_ccdf(color='r', ax=ax, label='data ccdf')
            fit.power_law.plot_ccdf(color='r', linestyle='--', ax=ax, label='fit ccdf')
            plt.legend()
            if not output_file:
                output_file = 'power_law_fit.png'
            fig.savefig(output_file)
        return fit # .power_law.alpha, fit.power_law.sigma 

    @staticmethod
    def create_igraph_object(nodes = None, edges = None, vertex_attrs = [], node_key = '', edge_attrs = []):
        """ create a Directed igraph Graph object from two dictionaries, the first
            representing the nodes in the graph and the second the edges. This function
            is meant to be used when each node and/or edge has some attributes, otherwise
            use the appropriate igraph function.
            The nodes and edges dictionaries are created using the two functions:
            read_data, read_edges

            You can pass either nodes or edges with or without attributes to this method.
            If nodes are given then no edges will be created. If edges are given then
            a full graph with vertices and edges will be created.
            But keep in mind:
            1) iGraph creates a contiguos set of vertices labels (indices) from edges.
               In most cases this is ok but if your edge list is created with large gaps
               among node labels the result will be wasteful.

            2) iGraph library returns an error if it founds a label in the edge set
               that is not in the node set. So if you pass both nodes and edges you must
               ensure that each label in the edge set is also in the node set.

            Args:
                nodes: (dict) 
                edges: (dict)
                vertex_attrs: (list) a list of all attributes of nodes
                node_key: (str) the key used to index the nodes dict
                edge_attrs: (list)   a list of all attributes of edges
        """

        print("create_igraph_object: creating Igraph Directed Graph object ...")
        if not bool(nodes) and not bool(edges):
            return Graph()

        eattrs = {}
        vattrs = {}

        if bool(nodes):
            #print("create_igraph_object: adding nodes and node attributes from input")
            N = len(nodes.keys())
            for el in vertex_attrs:
                vattrs.setdefault(el, [])
                for k in sorted(nodes.keys()): # WARNING: check this sort!
                    if el in nodes[k]:
                        vattrs[el].append(nodes[k][el])
                    elif el == node_key:
                        vattrs[el].append(k)
                    else:
                        vattrs[el].append('')
                #vattrs[el] = ['' if el not in nodes[k] else nodes[k][el] for k in nodes.keys()]
        else:
            N = 0

        if bool(edges):
            #print("create_igraph_object: adding edges and edge attributes from input")
            edgelist = [(e[0], e[1]) for e in edges]
            weight   = [1 for e in edges]
            eattrs['weight'] = weight
            if bool(edge_attrs):
                if len(edge_attrs) == 1:
                    eattrs[edge_attrs[0]] = [e[2] for e in edges]
                else:
                    for i in range(len(edge_attrs)):
                        eattrs[edge_attrs[i]] = [e[2][i] for e in edges]

        g = Graph(n = N, edges = edgelist, directed = True, edge_attrs = eattrs, vertex_attrs = vattrs) 

        return g

    @staticmethod
    def parse_igraph_object(graph=None, node_key = 'index'):
        """ create a aNodes object from an iGraph object. 
            Args:
                graph: the iGraph object 
                node_key: the vertex attribute to be used as key of the nodedict. if empty or not in graph.vs.attributes() iGraph indices will be used. note that node_key will NOT be in nodes_attributes
        """

        print("parse_igraph_object: creating aNodes object from Igraph Directed Graph object ...")
        
        if not graph:
            return aNodes()

        vattrs = graph.vs.attributes()
        if node_key in vattrs:
            keys = graph.vs[node_key]
            vattrs.remove(node_key)
        else:
            keys = graph.vs.indices
            print('Warning! Specified node key is not among the attributes, indices will be used')
            node_key = 'index'

        nodes = {key:{} for key in keys} 
        for att in vattrs:
            values = graph.vs[att]
            for i,key in enumerate(keys):
                nodes[key][att] = values[i]

        edges = []
        eattrs = graph.es.attributes()
        if len(eattrs) == 0:
            for e in graph.es:
                edges.append((e.source, e.target))
        elif len(eattrs) == 1:
            for e in graph.es:
                edges.append((e.source, e.target, e[eattrs[0]]))
        else:
            for e in graph.es:
                edges.append((e.source, e.target, [e[eattrs[i]] for i in range(len(eattrs))]))

        anodes = aNodes(nodedict=nodes, node_attributes=vattrs, node_key=node_key, edgelist=edges, edge_attributes=eattrs)
        
        return anodes
    
    def count(self, attribute='', cumulate = False, compute_stats = False):
        """ count the number of occurrences (number of nodes in nodedict) of each value for the given attribute 
            (using the Counter method).
            
            Args:
                attribute: the node attribute to count
                cumulate: compute the cumulate prefix sum of counts
                compute_stats: compute max, mean, median, std of the given attribute over
                               all the nodes, requires that the attribute values can be 
                               converted to int

            Return:
                counts: the number of nodes with the same value of the attribute 
                if compute_stats is True:
                    fill in self.stats:
                    stats['attribute'] = {max: value, mean: value, median: value, std: value}
        """
        if attribute == 'links':
            attribute = ['href', 'src', 'data', 'http', 'location']
            sat       = set(attribute)
            vals = []
            for k, v in self.nodedict.items():
                if sat.issubset(set(v.keys())):
                    vals.append(int(v['href']) + int(v['src']) + int(v['data']) + int(v['http']) + int(v['location']))
            attribute = 'links'
        else:
            vals = []
            for k, v in self.nodedict.items():
                if attribute in v.keys():
                    vals.append(v[attribute])

        counts = Counter(vals)
        #counts = counts.most_common()

        if compute_stats:
            stats = dict()
            vals  = np.array([int(v) for v in vals])
            stats['max']    = np.max(vals)
            stats['mean']   = np.mean(vals)
            stats['median'] = np.median(vals)
            stats['std']    = np.std(vals)
            self.stats[attribute] = stats

        if cumulate:
            cumul = [v/len(vals) for v in np.cumsum(counts)]
            print(cumul)

        return counts

    def dump_nodedict(self, out_file='stdout'):
        """ Dump a nested dictionary, items are printed in reverse alphabetical order (of the key)
        """
        if bool(out_file):
            if out_file != 'stdout':
                with open(out_file, 'w') as f:
                    for k, val in self.nodedict.items():
                        print("# key", *[v[0] for v in sorted(val.items(), reverse=True)], file=f)
                        break
                with open(out_file, 'a') as f:
                    for k, val in self.nodedict.items():
                        print(k, *[v[1] for v in sorted(val.items(), reverse=True)], file=f)
            else:
                for k, val in self.nodedict.items():
                    print(k, *[v[1] for v in sorted(val.items(), reverse = True)])
        return 

    def dump_edgelist(self, out_file='stdout'):
        """ Dump the edge list"""
        if bool(out_file):
            if out_file != 'stdout':
                with open(out_file, 'w') as f:
                    print("# {}".format(self.edge_attributes), file = f)

                with open(out_file, 'a') as f:
                    for e in self.edges:
                        print("{} {} {}".format(e[0], e[1], e[2]), file = f)
            else:
                for e in self.edges:
                    print("{} {} {}".format(e[0], e[1], e[2]))
        return 

    def dump_stats(self, out_file='stdout'):
        if bool(out_file):
            if out_file != 'stdout':
                with open(out_file, 'w') as f:
                    for k, val in self.stats.items():
                        print("# key", *[v[0] for v in sorted(val.items(), reverse=True)], file=f)
                        break
                with open(out_file, 'a') as f:
                    for k, val in self.stats.items():
                        print(k, *[v[1] for v in sorted(val.items(), reverse=True)], file=f)
            else:
                for k, val in self.stats.items():
                    print(k, *[v[1] for v in sorted(val.items(), reverse = True)])
        
        return


    def dump_table(self, attributes = [], n = 10, out_file = 'stdout', table_format = 'plain'):
        """
            print a sorted list of attributes given in the input list 'attributes'.
            The first field in the list must be a string representing the name of the node. 
        """
        val = {}
        default = {}

        idx = [v for v in self.nodedict.keys()]
        for i in attributes:
            default[i] = '' if type(self.nodedict[0][i]) is str else -1
        for i in attributes:
            val[i] = [self.nodedict[k][i] if i in self.nodedict[k] else default[i] for k in self.nodedict.keys()]

        siter = [v for v in zip(idx, *(val[i] for i in attributes))]
        b = {}
        headers = ['#id'] + attributes

        #Zero file
        if out_file != 'stdout':
            open(out_file, 'w')

            for i in range(2, len(headers)):
                s = sorted(siter, key = lambda x: x[i], reverse=True)[:n]
                # Set best self.nodedict
                for k in s:
                    b.setdefault(headers[i], [])
                    b[headers[i]].append(list(k)[0])
                # Dump a table
                if 1:
                    ch = list(headers)
                    ch[i] = ch[i].upper()
                    with open(out_file, 'a') as f:
                        print(tabulate(s, ch, tablefmt=table_format, numalign='left'), file = f)
                        print("Sorted {} degrees saved to file {}".format(ch[i], out_file))
        
        else:

            for i in range(2, len(headers)):
                s = sorted(siter, key = lambda x: x[i], reverse=True)[:n]
                # Set best self.nodedict
                for k in s:
                    b.setdefault(headers[i], [])
                    b[headers[i]].append(list(k)[0])
                # Dump a table
                if 1:
                    ch = list(headers)
                    ch[i] = ch[i].upper()
                    print(tabulate(s, ch, tablefmt=TAB_FORMAT, numalign='left'))
        return


    def acluster(self, attribute, value):
        """cluster nodes based on specified attribute"""
        n = self.nodedict
        d = {}
        if not bool(value):
            count = 0
            for k, v in n.items():
                if attribute in v:
                    d.setdefault(v[attribute], [])
                    d[v[attribute]].append(k)
                    count += 1

            d_nelems = count 

            membership = []
            count = 0
            for k, v in d.items():
                for i in range(len(v)):
                    membership.append(count)
                count += 1

            if (len(membership) != d_nelems):
                print("afilter: error calculating membership. Quit!")
                sys.exit()
        else:
            print("error, bool(value)=True currently not supported", file=sys.stderr)
            sys.exit(111) 

        return d, membership

    def iGraphCluster(self, attribute):
        """use acluster to define clusters and return an iGraph VertexClustering object which is ready for plot""" 
        # Filter nodes
        d, m = self.acluster(attribute, None)
        # Build a list of vertices from d
        vertices = list(itertools.chain(*d.values()))
        # Convert to igraph
        g        = self.iGraphObject()
        # Subgraph
        subg     = g.subgraph(vertices)
        # Use the subgraph and the membership array to 
        # construct a vertexclustering object
        cl = VertexClustering(subg, m)

        return cl

    def fit_degree(self, deg_type='alldeg', xmin=None, xmax=None, compare=False, plot=False, plot_file='', dump=False, dump_file=''):
        ''' fit degree distribution to power law
            
            Args:
                deg_type        the degree to fit; must be in ['outdeg', 'indeg', 'alldeg', 'outdegsimply', 'indegsimply'], default is 'alldeg'
                all other arguments are inherited from staticmethod fit_data
        '''
        if deg_type not in ['outdeg', 'indeg', 'alldeg', 'outdegsimply', 'indegsimply']:
            print('error: specified degree type is not supported. will use "alldeg"', file=sys.stderr)
            deg_type = 'alldeg'
        
        data = [v.get(deg_type,0) for _,v in self.nodedict.items()]
        fit = self.fit_data(data, xmin, xmax, compare, plot, plot_file)
        if dump:
            todump = {k:v for k,v in vars(fit).items() if v is None or isinstance(v,list) or isinstance(v,np.ndarray) or isinstance(v,int) or isinstance(v,float)}
            if not dump_file:
                dump_file = deg_type + '_power_law_fit_vars.pickle'
            with open(dump_file, 'wb') as f:
                pickle.dump(todump, f)
        return fit.alpha


    def read_graph(self, config_file, return_datastruct=False, return_both=False):
        """read options from config_file in .ini format, build an aNode object and,
        by default, return an iGraph object
        if return_datastruct is True, only return the aNode object
        if return_datastruct is False AND return_both is True, return both the iGraph 
        object and the aNodes object
        """
        config = configparser.ConfigParser()

        # Read nodelib configuration file
        if not os.path.isfile(config_file):
            print("{} is not a valid configuration file for nodelib. Quit.\n".format(config_file))
            sys.exit()
        config.read(config_file)

        # Read edges configuration
        edges_file       = config.get('EDGE', 'file')
        edge_attributes  = config.get('EDGE', 'attributes').split(',')
        edge_key         = config.get('EDGE', 'key')

        # Seems that the parser doesn't handle a None string
        if 'None' in edge_attributes:
            edge_attributes = None

        # At least the edge file must be set in config and be a regular file
        if edges_file == 'None' or edges_file == '': 
            print("read_graph: no edges_file in nodelib configuration! Quit.", file = sys.stderr)
            sys.exit()
        elif not os.path.isfile(edges_file):
            print("read_graph: file {} does not exists or is not a regular file".format(edges_file))
            sys.exit()
        else:
            print("read_graph: reading edges from {}, using attributes: {}".format(edges_file, edge_attributes))
        
        # Read the edge list from file and store it in "aNode" data struct
        p = self.getEdges(edges_file, edge_attributes = edge_attributes, node_key = edge_key)

        # Read nodes configuration, You can have several nodes file!
        NODES = list()
        for sec in config.sections():
            if 'NODE' in sec:
                NODES.append(sec)

        # For each node section get the options and see if there is a regular node file
        # to read.
        for nodesec in NODES:
            nfile            = config.get(nodesec, 'file')
            print("NFILE = ", nfile)
            if nfile != 'None' or nfile == '':
                if os.path.isfile(nfile):
                    nattributes      = config.get(nodesec, 'attributes').split(',')
                    nkey             = config.get(nodesec, 'key')
                    nattributes      = [a.strip() for a in nattributes]

                    # Seems that the parser doesn't handle a None string
                    if 'None' in nattributes:
                        nattributes = None

                    # Read node with attributes and store them in the aNode data struct
                    print("read_graph: reading nodes from {}, using key: {}, using attributes: {}".format(nfile, nkey, nattributes))
                    print("(oldkey) p.node_key = ", p.node_key)
                    if nkey != p.node_key:
                        #print("read_graph: find a new key!")
                        p.addAttributesFromFile(nfile, nattributes, nkey, p.node_key)
                    else:
                        p.addAttributesFromFile(nfile, nattributes, nkey, p.node_key)
                else:
                    print("read_graph: ERROR file {} does not exist or is not a regular file.".format(nfile))
                    sys.exit(111)
                    

        # Read the vector file and attributes from config
        vfile = config.get('VECTOR', 'file')

        if vfile != 'None' or vfile == '':
            if os.path.isfile(vfile):
                vattributes      = config.get('VECTOR', 'attributes').split(',')
                vattr_from_file  = config.get('VECTOR', 'attributes_from_file')

                vattributes = [a.strip() for a in vattributes]
                # Seems that the parser doesn't handle a None string
                if 'None' in vattributes:
                    vattributes = None

                # Read vectors from file
                p.vectordict, p.vector_categories = p.read_vectors(vfile, categories_from_file = vattr_from_file)
                print("read_graph: read vector from {}, categories: {}".format(vfile, p.vector_categories))

                if not vattr_from_file and bool(vattributes):
                    p.vector_categories = vattributes

        # Read the nodelabel section
        nlabel = config.get('LABELS', 'label')
        nname  = config.get('LABELS', 'name')

        if 'None' in nlabel:
            nlabel = None
        if 'None' in nname:
            nname = None


        print("==================== anode datastruct ====================")
        print("Build data struct with {} nodes and {} edges".format(len(p.nodedict), len(p.edges)))
        print("EDGE ATTRIBUTES: {}".format(p.edge_attributes))
        print("NODE ATTRIBUTES: {}".format(p.node_attributes))
        print("==================== ================ ====================")

        if return_datastruct == True:
            return p


        # ===================================================================== 
        # At the very end, we can create the graph object
        g = p.iGraphObject()
        print("Graph Obj, node attributes: {}".format(g.vs.attributes()))

        # Set the label attribute
        if bool(nlabel):
            g.vs['label'] = g.vs[nlabel]
        if bool(nname):
            g.vs['name']  = g.vs[nname]
        # ===================================================================== 


        # DEBUG PRINT
        DEBUG_PRINT = False
        if (DEBUG_PRINT):
            print("--"*70)
            print(p.node_attributes)
            for k, v in p.nodedict.items():
                print(k, *v)
            print(p.nodedict)
            print("--"*70)
            print(p.edge_attributes)
            print(p.edges)
            print("--"*70)
            print(p.vector_categories)
            print(p.vectordict)
            print("--"*70)
            print(g)
            print("--"*70)
            for v in g.vs:
                print(v)
        
        if return_both == True:
            return p, g

        return g

# Read graph using config.ini
def main1():
    if len(sys.argv) < 2:
        print("usage: {} config.ini".format(sys.argv[0]))
        sys.exit()

    cfile = sys.argv[1]
    print("Using config file: {}".format(cfile))
    
    an = aNodes()
    g = an.read_graph(cfile)

    print(g.summary())

# Read graphs without config.ini
def main2():


    DEFAULT_EDGES_ATTR  = ['label']
    DEFAULT_NODES_ATTR  = ['nodeId', 'url', 'code', 'record', 'href', 'src', 'data', 'http', 'location']
    DEFAULT_NODES_STAT  = ['record', 'langTika', 'langDet', 'wordCount']
    DEFAULT_KEY         = 'record'
    DEFAULT_VECTOR_CAT  = ['Social Network', 'Information System', 'Fraud', 'Gambling', 'Murder and grievous bodily injury', 'Sexual exploitation of children and child pornography', 'Illicit trafficking in weapons and munitions and explosives', 'Rape', 'Racism and xenophobia', 'IT Services companies', 'Media companies', 'Terrorism', 'Weaponary', 'Illicit drugs', 'Cyber Attack', 'Cyber deception', 'Cyber Security'] 
    # g = aNodes.dataStruct(nodedict={'a':1})
    # f = aNodes(nodedict={'b':1})

    # print(g)
    # print(g.nodedict)
    # print(f)
    # print(f.nodedict)


    nodes_file   = sys.argv[1]
    stats_file   = sys.argv[2]
    vectors_file = sys.argv[3]
    edges_file   = sys.argv[4]

    # h = aNodes.getNodes(nodes_file, DEFAULT_NODES_ATTR, DEFAULT_KEY)
    # h.addAttributesFromFile(stats_file, DEFAULT_NODES_STAT, DEFAULT_KEY)
    #print(h)
    #print(h.nodedict)

    # i = aNodes.getVectors(vectors_file, vector_categories = [], categories_from_file = False)
    # print(i.vector_categories)
    # print(i.vectordict)
    # print(i.nodedict)

    # i.addAttributesFromFile(nodes_file, DEFAULT_NODES_ATTR, DEFAULT_KEY)
    # print(i.nodedict)
    # i.addAttributesFromFile(stats_file, DEFAULT_NODES_STAT, DEFAULT_KEY)
    # print(i.nodedict)

    p = aNodes.getEdges(edges_file, edge_attributes = DEFAULT_EDGES_ATTR)
    #print(p.edges)

    p.addAttributesFromFile(nodes_file, DEFAULT_NODES_ATTR, 'nodeId')
    #print(p.nodedict)

    p.addAttributesFromFile(stats_file, DEFAULT_NODES_STAT, 'record', 'nodeId')
    for k,v in p.nodedict.items():
        print(k, v)
    #print(p.nodedict)

    print('')
    print('Subgraph with langDet == it:')
    s = p.getSubgraphByAttribute('langDet', 'it')
    for k,v in s.nodedict.items():
        print(k, v)
    # print(s.edges)

    #g = p.iGraphObject()
    #for v in g.vs:
    #    print(v)
    #print(g.summary())

    

    return 0


if __name__ == '__main__':
    main1()

